/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: phg-mpi.c,v 1.47 2012/04/05 08:03:16 zlb Exp $ */

#include "phg.h"

#if USE_ZOLTAN
#include "partition-zoltan.h"
#endif	/* USE_ZOLTAN */

GRID *_phg_g_sys = NULL;

int phgRank = 0, phgNProcs = 1, phgCmdSerialno = 0;
MPI_Comm phgComm = MPI_COMM_WORLD;

#if USE_MPI

#include <stdlib.h>

/* for getpid() */
#include <sys/types.h>
#include <unistd.h>
#include <limits.h>	/* PATH_MAX */

static int functions_count = 0;
static PFunc_t *functions = NULL;
static int mpi_initialized = FALSE;

#if DEBUG
char (*phgProcessorNames)[MPI_MAX_PROCESSOR_NAME + 1] = NULL;

int _phg_comm_count = 0, _phg_comm_count_max = 0;
#undef MPI_Comm_dup
#undef MPI_Comm_create
#undef MPI_Comm_free
int (*MPI_Comm_split_real)(MPI_Comm comm, int color, int key, MPI_Comm *commp);
int (*MPI_Comm_dup_real)(MPI_Comm comm, MPI_Comm *comm_out);
int (*MPI_Comm_create_real)(MPI_Comm comm, MPI_Group group, MPI_Comm *commp);
int (*MPI_Comm_free_real)(MPI_Comm *commp);
#endif	/* DEBUG */

MPI_Comm
phgInitSetComm(MPI_Comm comm)
/* sets up the world communicator (must be called before phgInit()).
 * returns old setting. */
{
    MPI_Comm old_comm;

    if (phgInitialized)
	phgError(1, "%s must be called before phgInit().\n", __func__);
    old_comm = phgComm;
    phgComm = comm;

    return old_comm;
}

int
phgRegisterParallelFunction(PFunc_t func, const char *funcname,
				const char *file, int line)
{
    if (phgInitialized) {
	phgInfo(-1, "%s:%d, unregistered parallel function called.\n",
		file, line);
	phgInfo(-1, "Please add the following line:\n");
	phgInfo(-1, "    RegisterParallelFunction(%s);\n", funcname);
	phgInfo(-1, "before phgInit(...).\n");
	phgError(1, "abort.\n");
    }

    phgInfo(3, "registering parallel function %d: %s\n",
	functions_count, funcname);
    functions = phgRealloc_(functions, sizeof(PFunc_t) * (functions_count + 1),
					sizeof(PFunc_t) * functions_count);
    functions[functions_count] = func;

    return functions_count++;
}

MPI_Op		MPI_MSM;
MPI_Datatype	MPI_3DOUBLE;

/* User function for MPI_Op_create which computes min, sum and max */
static void
MSM(void *in, void *inout, int *len, MPI_Datatype *type)
/* computes b := a op b */
{
    double *a = in, *b = inout;
    int i;

    assert(*type == MPI_3DOUBLE);
    for (i = 0; i < *len; i++) {
	b[0] = (b[0] < a[0] ? b[0] : a[0]); 
	b[1] = a[1] + b[1];
	b[2] = (b[2] > a[2] ? b[2] : a[2]); 
	b += 3;
	a += 3;
    }
}

#if FT_PHG == FT___FLOAT128
MPI_Op		MPI_OP_SUM, MPI_OP_MAX;
MPI_Datatype	MPI_FLOAT128;

/* User function for MPI_OP_SUM */
static void
__sum(void *in, void *inout, int *len, MPI_Datatype *type)
/* computes b := a op b */
{
    FLOAT *a = in, *b = inout;
    int i;

    assert(*type == MPI_FLOAT128);
    for (i = 0; i < *len; i++)
	*(b++) += *(a++);
}

/* User function for MPI_OP_MAX */
static void
__max(void *in, void *inout, int *len, MPI_Datatype *type)
/* computes b := a op b */
{
    FLOAT *a = in, *b = inout;
    int i;

    assert(*type == MPI_FLOAT128);
    for (i = 0; i < *len; i++, a++, b++)
	if (*b < *a)
	    *b = *a;
}
#endif	/* FT_PHG == FT___FLOAT128 */

static BOOLEAN trap_mpi_error = FALSE;
static MPI_Errhandler errhandler = MPI_ERRHANDLER_NULL;

static void
errhandler_function(MPI_Comm *comm, int *code, ...)
{
    int len;
    char errmsg[MPI_MAX_ERROR_STRING + 1];

    MPI_Error_string(*code, errmsg, &len);
    phgError(1, "MPI error: %s\n", errmsg);
}

static void
finalize_mpi(GRID *g)
{
    Unused(g);

    ParallelFunction(finalize_mpi, TRUE);

    if (functions != NULL) {
	phgFree(functions);
	functions = NULL;
	functions_count = 0;
    }

    if (phgNProcs <= 0)
	return;

    MPI_Op_free(&MPI_MSM);
    MPI_Type_free(&MPI_3DOUBLE);

#if FT_PHG == FT___FLOAT128
    MPI_Op_free(&MPI_OP_SUM);
    MPI_Op_free(&MPI_OP_MAX);
    MPI_Type_free(&MPI_FLOAT128);
#endif	/* FT_PHG == FT___FLOAT128 */

#if !NO_NEW_COMMUNICATOR
    MPI_Comm_free(&phgComm);
#endif	/* !NO_NEW_COMMUNICATOR */

    if (errhandler != MPI_ERRHANDLER_NULL)
	MPI_Errhandler_free(&errhandler);

    if (!mpi_initialized) {
	MPI_Finalize();
    }
    else {
	/* to prevent slave processes from exiting */
	phgRank = 0;
    }

    if (phgMasterSlave && phgRank > 0)
	exit(0);

    phgNProcs = -1;

    return;
}

void
phgInitMPI(int *argc, char **argv[])
{
    char name[MPI_MAX_PROCESSOR_NAME + 1], cwd[PATH_MAX + 1];
    int len;
    static BOOLEAN first_call = TRUE;

    /* The first call is just for registering MPI-related options */
    if (first_call) {
	first_call = FALSE;
	phgOptionsRegisterNoArg("-trap_mpi_error",
				"Use custom MPI error handler",
				&trap_mpi_error);
	return;
    }

#if DEBUG
    MPI_Comm_split_real	= MPI_Comm_split;
    MPI_Comm_dup_real	= MPI_Comm_dup;
    MPI_Comm_create_real= MPI_Comm_create;
    MPI_Comm_free_real	= MPI_Comm_free;
#endif	/* DEBUG */

    MPI_Initialized(&mpi_initialized);
    /* save cwd */
    if (getcwd(cwd, sizeof(cwd)) == NULL)
	phgError(1, "can't determine cwd, abort.\n");
    if (!mpi_initialized) {
	assert(phgComm == MPI_COMM_WORLD);
	if (MPI_Init(argc, argv))
	    phgError(1, "MPI_Init failed, abort.\n");
    }

    /* make a duplicate of phgComm */
#if !NO_NEW_COMMUNICATOR
    MPI_Comm_dup(phgComm, &phgComm);
#endif	/* !NO_NEW_COMMUNICATOR */

    /* broadcast and restore cwd */
    MPI_Bcast(cwd, sizeof(cwd), MPI_BYTE, 0, phgComm);
    if (chdir(cwd) == -1)
	phgError(1, "can't change to directory \"%s\", abort.\n", cwd);

    MPI_Comm_rank(phgComm, &phgRank);
    MPI_Comm_size(phgComm, &phgNProcs);

    /* Initialize parallel functions */
    RegisterParallelFunction(finalize_mpi);
    RegisterParallelFunction(phgSetVerbosity_);
    RegisterParallelFunction(phgNewGrid_);
    RegisterParallelFunction(phgPartitionGrid__);
    RegisterParallelFunction(phgRedistributeGrid);
    RegisterParallelFunction(phgRefineMarkedElements);
    RegisterParallelFunction(phgCoarsenMarkedElements);
    RegisterParallelFunction(phgRefineAllElements_);
    RegisterParallelFunction(phgRefineRandom_);
    RegisterParallelFunction(phgRefineSurface);
    RegisterParallelFunction(phgCheckConformity);
    RegisterParallelFunction(phgFreeGrid_);
    RegisterParallelFunction(phgDumpGridInfo);
    RegisterParallelFunction(phgDumpGrid);
    RegisterParallelFunction(phgDumpTree);
    RegisterParallelFunction(phgExportVTK_);
    RegisterParallelFunction(phgExportDX_);
    RegisterParallelFunction(phgExportEnsight_);

    MPI_Get_processor_name(name, &len);
    name[len] = '\0';
#if DEBUG
    len = sizeof(phgProcessorNames[0]);
    phgProcessorNames = phgAlloc(phgNProcs * len);
    MPI_Allgather(name, len, MPI_BYTE, phgProcessorNames, len, MPI_BYTE,
		  phgComm);
#endif

    MPI_Type_contiguous(3, MPI_DOUBLE, &MPI_3DOUBLE);
    MPI_Type_commit(&MPI_3DOUBLE);
    MPI_Op_create(MSM, 1, &MPI_MSM);

#if FT_PHG == FT___FLOAT128
    MPI_Type_contiguous(SIZEOF_PHG_FLOAT, MPI_BYTE, &MPI_FLOAT128);
    MPI_Type_commit(&MPI_FLOAT128);
    MPI_Op_create(__sum, 1, &MPI_OP_SUM);
    MPI_Op_create(__max, 1, &MPI_OP_MAX);
#endif	/* FT_PHG == FT___FLOAT128 */

    phgOptionsParseCmdline(argc, argv);

    /* Initialize modules after parsing all options.
     * Note: also add the calls below in phgInit() in utils.c */
    phgPerfInit();			/* init phgPerf */
    phgSolverInitialize(argc, argv);	/* init phgSolver */
#if USE_ZOLTAN
    phgPartitionZoltanInitialize(*argc, *argv);
#endif	/* USE_ZOLTAN */

    phgInitialized = TRUE;
    phgOptionsHelp();

    if (phgRank == 0) {
	phgPrintf("Parallel Hierarchical Grid (version " PHG_VERSION ").\n");
	fflush(stdout);
	if (phgVerbosity >= 2) {
	    phgPrintf("sizeof(INT)        = %d\n", sizeof(INT));
	    phgPrintf("sizeof(FLOAT)      = %d\n", sizeof(FLOAT));
	    phgPrintf("sizeof(SIMPLEX)    = %d\n", sizeof(SIMPLEX));
	    phgPrintf("sizeof(GRID)       = %d\n", sizeof(GRID));
	    phgPrintf("sizeof(RNEIGHBOUR) = %d\n", sizeof(RNEIGHBOUR));
	}
	if (phgVerbosity >= 1) {
	    phgPrintf("number of process(es): %d\n", phgNProcs);
	    phgPrintf("%s mode.\n", phgMasterSlave ? "Master-slave" : "SPMD");
	}
	fflush(stdout);
    }
    phgInfo(1, ">>> processor=%s, pid=%d <<<\n", name, getpid());

    phgTrapSignals_(FALSE);
    phgTimerInit(FALSE);

    if (trap_mpi_error) {
	/* Create custom error handler to facilitate MPI error debugging */
#if 0	/* MPI-2 */
	MPI_Comm_create_errhandler(errhandler_function, &errhandler);
	MPI_Comm_set_errhandler(phgComm, errhandler);
#else
	MPI_Errhandler_create(errhandler_function, &errhandler);
	MPI_Errhandler_set(phgComm, errhandler);
#endif
    }

    if (phgRank > 0 && phgMasterSlave) {
	if (_phg_pause)
	    phgPause(0);
	phgSlave();
    }

    return;
}

void
phgFinalizeMPI(void)
{
    finalize_mpi(_phg_g_sys);

#if DEBUG
    phgFree(phgProcessorNames);
    phgProcessorNames = NULL;
#endif
}

void
phgCallSlaves(int cmd)
{
    if (phgMasterSlave) {
	phgInfo(3, "bringing slave to function %d\n", cmd);
	MPI_Bcast(&cmd, 1, MPI_INT, 0, phgComm);
	phgCmdSerialno++;
    }
}

void
phgSlave(void)
{
    int cmd;

    while (TRUE) {
	/* wait for next command */
	phgInfo(3, "phgSlave: waiting for command.\n");
	MPI_Bcast(&cmd, 1, MPI_INT, 0, phgComm);
	if (cmd >= functions_count) 
	    phgError(1, "phgSlave: invalid command %d\n", cmd);
	phgCmdSerialno++;
	phgInfo(3, "entering function %d, g_sys=%p\n", cmd, _phg_g_sys);
	functions[cmd](_phg_g_sys);
	phgInfo(3, "returning from function %d, g_sys=%p\n", cmd, _phg_g_sys);
    }
}

#endif /* USE_MPI */
