/* $Id: cut_test.c,v 1.23 2012/07/22 07:59:21 zlb Exp $ */

#include "phg.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define PLANE_CUT	0	/* 0 ==> sphere cut */

static FLOAT
cut_plane(FLOAT x, FLOAT y, FLOAT z)
/* The equation for the cut plane */
{
#if PLANE_CUT
    return 1.0 * (x - 0.5) + 2.0 * (y - 0.5) + 3.0 * (z - 0.5);
#else	/* PLANE_CUT */
    return Pow(x - 0.5, 2) + Pow(y - 0.5, 2) + Pow(z - 0.5, 2) - 0.25 * 0.25;
#endif	/* PLANE_CUT */
}

void
cut_plane_marker(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
/* The marker function */
{
    *value = (cut_plane(x, y, z) > 0.0 ? 1.0 : -1.0);
}

static void
func(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    *value = x * x + y * y + z * z;
}

int
main(int argc, char *argv[])
{
    GRID *g, *g1;
    const char *fn = "cube.dat";
    SIMPLEX *e;
    DOF *u0, *u, *v;
    FLOAT a, b, t, *f, *cuts;
    INT refines = 1;
    double time;
    INT i, count;
    int k, v0, v1;

    phgOptionsRegisterInt("refines", "Number of refinements", &refines);
    phgInit(&argc, &argv);

    g = phgNewGrid(-1);
    if (argc > 1)
	fn = argv[1];
    if (!phgImport(g, fn, FALSE))
	phgError(1, "can't read file \"%s\".\n", fn);

    g1 = NULL;
    for (k = 0; k < refines && g->nelem_global < 200000 * g->nprocs; k++) {
	VEF_MAP *m;
	phgRefineAllElements(g, 1);
	/*phgRefineRandomElements(g, "50%");*/
	phgBalanceGrid(g, 1.0, 1, NULL, 1.0);

	cuts = phgAlloc(g->nedge * sizeof(*cuts));
	f = phgAlloc(g->nvert * sizeof(*f));
	for (i = 0; i < g->nvert; i++)
	    f[i] = cut_plane(g->verts[i][0], g->verts[i][1], g->verts[i][2]);
	m = phgDofSetupVEFMap(g, NULL, EDGE_FLAG);
	for (count = i = 0; i < g->nedge; i++) {
	    cuts[i] = -1.0;
	    e = m->Emap[i];
	    if (e == NULL)
		continue;	/* unreferenced edge */
	    GetEdgeVertices(e, m->Eind[i], v0, v1);
	    a = f[e->verts[v0]];
	    b = f[e->verts[v1]];
	    if (a == 0. || b == 0.)
		continue;
	    if ((a >= 0.) != (b >= 0.)) {
#if PLANE_CUT
		t = a / (a - b);
#else	/* PLANE_CUT */
		/* find t by bisection */
		FLOAT x0 = g->verts[e->verts[v0]][0],
		      y0 = g->verts[e->verts[v0]][1],
		      z0 = g->verts[e->verts[v0]][2],
		      x1 = g->verts[e->verts[v1]][0],
		      y1 = g->verts[e->verts[v1]][1],
		      z1 = g->verts[e->verts[v1]][2];
		FLOAT t0 = 0.0, t1 = 1.0, x, y, z, c;

		t = 0.;
		while (t1 - t0 > 1e-6) {
		    t = 0.5 * (t0 + t1);
		    c = cut_plane(x = 0.5 * (x0 + x1),
				  y = 0.5 * (y0 + y1),
				  z = 0.5 * (z0 + z1));
		    if (c == 0.)
			break;
		    if ((c > 0.) == (a > 0.)) {
			t0 = t;
			a = c;
			x0 = x;
			y0 = y;
			z0 = z;
		    }
		    else {
			t1 = t;
			b = c;
			x1 = x;
			y1 = y;
			z1 = z;
		    }
		}
#endif	/* PLANE_CUT */
		if (t > 1e-6 && t < (1.0 - 1e-6)) {
		    cuts[i] = t;
		    if (g->types_edge[i] & OWNER)
			count++;
		}
	    }
	}
#if 0
	for (i = 0; i < g->nedge; i++)
	    phgPrintf("cuts[%d] = %le\n", i, cuts[i]);
#endif
#if USE_MPI
	if (g->nprocs > 1) {
	    i = count;
	    MPI_Reduce(&i, &count, 1, PHG_MPI_INT, MPI_SUM, 0, g->comm);
	}
#endif

	phgPrintf("\nTotal number of cuts: %"dFMT"\n", count);
	time = phgGetTime(NULL);
	phgFreeGrid(&g1);
	g1 = phgSurfaceCut(g, cuts);
	phgPrintf("%d+%d vertices, %d+%d elements, wall time: %lf.\n",
		g->nvert_global, g1->nvert_global - g->nvert_global,
		g->nelem_global, g1->nelem_global - g->nelem_global,
		phgGetTime(NULL) - time);
		
	phgFree(cuts);
	phgFree(f);
	phgDofFreeVEFMap(&m);

	if (phgVerbosity > 0) {
	    phgInfo(-1, "g1->nvert=%d, g1->nvert_global=%d\n",
			g1->nvert, g1->nvert_global);
	    phgInfo(-1, "g1->nedge=%d, g1->nedge_global=%d\n",
			g1->nedge, g1->nedge_global);
	    phgInfo(-1, "g1->nface=%d, g1->nface_global=%d\n",
			g1->nface, g1->nface_global);
	    phgInfo(-1, "g1->nelem=%d, g1->nelem_global=%d\n",
			g1->nelem, g1->nelem_global);
	    phgInfo(-1, "%d+%d vertices, %d+%d elements.\n", 
			g->nvert_global, g1->nvert_global - g->nvert_global,
			g->nelem_global, g1->nelem_global - g->nelem_global);
	}

#if 0
	if (g1->nelem <100) {
	    ForAllElements(g, e)
		phgPrintf("Element %d in g ==> Element %d in g1\n",
					e->index, g->alien_map[e->index]);
	    phgPrintf("\n");
	    ForAllElements(g1, e)
		phgPrintf("Element %d in g1 ==> Element %d in g\n",
					e->index, g1->alien_map[e->index]);
	}
#endif

	phgCheckConformity(g1);
	u0 = phgDofNew(g, DOF_DEFAULT, 1, "u", func);

	/* test phgDofTransfer */
	phgPrintf("    Testing phgDofTransfer:\n");
	time = phgGetTime(NULL);
	phgPrintf("\tg ==> g1, ");
	v = phgDofTransfer(u0, g1);
	phgPrintf("time = %0.4lg\n", phgGetTime(NULL) - time);
	time = phgGetTime(NULL);
	phgPrintf("\tg1 ==> g, ");
	u = phgDofTransfer(v, g);
	phgPrintf("wall time = %0.4lg\n", phgGetTime(NULL) - time);
	phgDofAXPY(-1.0, u0, &u);
	phgPrintf("\terror=%le, norm=%le\n", phgDofNormL2(u), phgDofNormL2(u0));
	phgDofFree(&u);
	phgDofFree(&v);

	/* test phgInterGridDofTransfer */
	phgPrintf("    Testing phgInterGridDofTransfer:\n");
	u = phgDofNew(g, DOF_DEFAULT, 1, "u", DofNoAction);
	v = phgDofNew(g1, DOF_DEFAULT, 1, "v", DofNoAction);
	time = phgGetTime(NULL);
	phgPrintf("\tg ==> g1, ");
	phgInterGridDofTransfer(u0, v);
	phgPrintf("time = %0.4lg\n", phgGetTime(NULL) - time);
	time = phgGetTime(NULL);
	phgPrintf("\tg1 ==> g, ");
	phgInterGridDofTransfer(v, u);
	phgPrintf("wall time = %0.4lg\n", phgGetTime(NULL) - time);
	phgDofAXPY(-1.0, u0, &u);
	phgPrintf("\terror=%le, norm=%le\n", phgDofNormL2(u), phgDofNormL2(u0));
	phgDofFree(&u);
	phgDofFree(&v);

	phgDofFree(&u0);
    }

    /* save the grid (with a marker function) */
    if (g1 != NULL) {
	u = phgDofNew(g1, DOF_P0, 1, "marker", cut_plane_marker);
	phgExportVTK(g1, "tmp.vtk", u, NULL);
	phgDofFree(&u);
    }

    phgFreeGrid(&g1);
    phgFreeGrid(&g);
    phgFinalize();

    return 0;
}
